/*
 * Higher-level operations on arrays -
 * intersect, sharp, absorb, etc.
 */
#include <stdio.h>
#include "logic.h"

/*
 * Check if cube p subsumes cube q
 */
subsume(p, q, nv)
struct cube *p, *q;
int nv;
{
	int i;
	register char *a, *b;

	a = p->c_logic;
	b = q->c_logic;
	for( i = nv; i > 0; --i )
		if( (*a++ & ~*b++) != 0 )
			return 0;
	return 1;
}

/*
 * Absorb an array
 */
absorb(a)
register struct array *a;
{
	struct cube **p0, *p, **q0, *q, *l;

	l = NULL;
	for( p0 = &a->a_first; (p = *p0) != NULL; ){
		for( q0 = &p->c_next; (q = *q0) != NULL; ){
			if( subsume(q, p, a->a_nvar) ){
				/* delete cube q */
				*q0 = q->c_next;
				cfree(q);
			} else if( subsume(p, q, a->a_nvar) )
				break;
			else
				q0 = &q->c_next;
		}
		if( q != NULL ){
			/* delete cube p */
			*p0 = p->c_next;
			cfree(p);
		} else {
			l = p;
			p0 = &p->c_next;
		}
	}
	a->a_last = l;
}

/* Find the intersection of 2 cubes - return 0 if they don't intersect */
cubix(r, a, b, nv)
register struct cube *r, *a, *b;
int nv;
{
	int i;

	for( i = 0; i < nv; ++i )
		if( (r->c_logic[i] = a->c_logic[i] & b->c_logic[i]) == N )
			return 0;	/* no intersection */
	return 1;
}

/*
 * Find the intersection of a cube with an array.
 */
struct array *
axc(a, c)
register struct array *a;
struct cube *c;
{
	struct array *b;
	struct cube *p, *n;
	int i;

	b = newarray(a->a_noin, a->a_nvar);
	n = newcube(a->a_nvar);
	for( p = a->a_first; p != NULL; p = p->c_next ){
		if( cubix(n, p, c, a->a_nvar) ){
			/* cube intersects - add it in */
			add(b, n);
			n = newcube(a->a_nvar);
		}
	}
	cfree(n);
	return b;
}

/*
 * Find the intersection of 2 arrays.
 */
struct array *
intersect(a, b)
struct array *a, *b;
{
	struct array *c;
	struct cube *p;

	arr_compat(a, b);
	c = newarray(a->a_noin, a->a_nvar);
	for( p = b->a_first; p != NULL; p = p->c_next ){
		append(c, axc(a, p));
		absorb(c);
	}
	return c;
}

/*
 * Sharp a cube from an array.
 */
struct array *
ashc(a, c)
register struct array *a;
struct cube *c;
{
	struct array *b;
	struct cube *p, *d, *x;
	int i, j;

	b = newarray(a->a_noin, a->a_nvar);
	x = newcube(a->a_nvar);
	for( p = a->a_first; p != NULL; p = p->c_next ){
	    if( !cubix(x, p, c, a->a_nvar) ){
		/* no intersection - copy the cube */
		d = newcube(a->a_nvar);
		ccopy(d, p, a->a_nvar);
		add(b, d);
	    } else {
		for( i = 0; i < a->a_nvar; ++i ){
		    if( p->c_logic[i] == X && x->c_logic[i] != X ){
			d = newcube(a->a_nvar);
			ccopy(d, p, a->a_nvar);
			d->c_logic[i] = X - x->c_logic[i];
			add(b, d);
		    }
		}
	    }
	}
	cfree(x);
	absorb(b);
	return b;
}

/*
 * Sharp array b from array a.
 */
struct array *
sharp(a, b)
register struct array *a, *b;
{
	struct array *c;
	struct cube *p;

	arr_compat(a, b);
	if( b->a_first == NULL )
		return acopy(a);
	p = b->a_first;
	c = ashc(a, p);
	while( (p = p->c_next) != NULL && c->a_first != NULL )
		hass(c, ashc(c, p));
	return c;
}

/*
 * Calculate the cost of a cube -
 * i.e. the number of inputs which aren't X.
 */
cost(c, ni)
struct cube *c;
int ni;
{
	int i, t;
	register char *s;

	t = 0;
	s = c->c_logic;
	for( i = 0; i < ni; ++i )
		if( *s++ != X )
			++t;
	return t;
}

/* Split on column i */
split(pi, z, o, x, i)
struct array *pi;
struct array **z, **o, **x;
int i;
{
	struct array *zeroes, *ones, *exes;
	struct cube *p, *q;

	zeroes = newarray(pi->a_noin, pi->a_nvar);
	ones = newarray(pi->a_noin, pi->a_nvar);
	exes = newarray(pi->a_noin, pi->a_nvar);
	for( p = pi->a_first; p != NULL; p = q ){
		q = p->c_next;
		switch( p->c_logic[i] ){
		case Z:		add(zeroes, p); break;
		case O: 	add(ones, p);	break;
		case X: 	add(exes, p);	break;
		default:	cfree(p);
		}
		p->c_logic[i] = X;
	}
	pi->a_first = NULL;
	*z = zeroes;
	*o = ones;
	*x = exes;
}

/* Make a universe array */
struct array *
univ_array(noin, nvar)
int noin, nvar;
{
	struct array *univ;
	struct cube *p;

	univ = newarray(noin, nvar);
	p = newcube(nvar);
	setcube(p, 0, nvar, X);
	add(univ, p);
	return univ;
}

/*
 * Check that the array a is a valid truth table and add don't care's.
 * A valid truth table is one for which the output specifications don't
 * intersect for any combination of inputs.
 * At present, we assume there is only one output.
 */
check_tt(a, onum)
struct array *a;
int onum;
{
	struct array *zeroes, *ones, *exes;
	struct array *zo, *ox, *zx, *u;

	split(a, &zeroes, &ones, &exes, a->a_noin);
	zo = intersect(zeroes, ones);
	ox = intersect(ones, exes);
	zx = intersect(zeroes, exes);
	if( zo->a_first != NULL ){
		yyerror("conflict between 0's and 1's for output %d on:",
			onum+1);
		prary(zo, stderr);
	}
	if( ox->a_first != NULL ){
		yyerror("conflict between 1's and %%'s for output %d on:",
			onum+1);
		prary(ox, stderr);
	}
	if( zx->a_first != NULL ){
		yyerror("conflict between 0's and %%'s for output %d on:",
			onum+1);
		prary(zx, stderr);
	}
	release(exes);
	append(zeroes, acopy(ones));
	u = univ_array(a->a_noin, a->a_nvar);
	exes = sharp(u, zeroes);
	release(zeroes);
	release(u);
	chcol(ones, a->a_noin, Z, O, O);
	append(ones, exes);
	hass(a, ones);
}

/* Invert the function represented by array a.
   At present, we assume that there is only one output. */
struct array *
inv_array(a)
struct array *a;
{
	struct array *b, *on, *off, *dc, *u;

	b = acopy(a);
	split(b, &off, &on, &dc, a->a_noin);
	release(off);
	append(on, acopy(dc));
	u = univ_array(a->a_noin, a->a_nvar);
	hass(on, sharp(u, on));
	release(u);
	chcol(on, a->a_noin, Z, O, O);
	append(on, dc);
	return on;
}
